"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const listingTools_1 = require("@blockworks-foundation/mango-v4-settings/lib/helpers/listingTools");
const anchor_1 = require("@coral-xyz/anchor");
const anchor_2 = require("@project-serum/anchor");
const spl_governance_1 = require("@solana/spl-governance");
const web3_js_1 = require("@solana/web3.js");
const fs_1 = __importDefault(require("fs"));
const oracle_1 = require("../src/accounts/oracle");
const builder_1 = require("../src/builder");
const client_1 = require("../src/client");
const clientIxParamBuilder_1 = require("../src/clientIxParamBuilder");
const constants_1 = require("../src/constants");
const utils_1 = require("../src/utils");
const constants_2 = require("./governanceInstructions/constants");
const createProposal_1 = require("./governanceInstructions/createProposal");
const voteStakeRegistryClient_1 = require("./governanceInstructions/voteStakeRegistryClient");
const { MB_CLUSTER_URL, PROPOSAL_TITLE, VSR_DELEGATE_KEYPAIR, VSR_DELEGATE_FROM_PK, DRY_RUN, } = process.env;
const getApiTokenName = (bankName) => {
    if (bankName === 'ETH (Portal)') {
        return 'ETH';
    }
    return bankName;
};
async function buildClient() {
    return await client_1.MangoClient.connectDefault(MB_CLUSTER_URL);
}
async function setupWallet() {
    const clientKeypair = web3_js_1.Keypair.fromSecretKey(Buffer.from(JSON.parse(fs_1.default.readFileSync(VSR_DELEGATE_KEYPAIR, 'utf-8'))));
    const clientWallet = new anchor_1.Wallet(clientKeypair);
    return clientWallet;
}
async function setupVsr(connection, clientWallet) {
    const options = anchor_1.AnchorProvider.defaultOptions();
    const provider = new anchor_1.AnchorProvider(connection, clientWallet, options);
    const vsrClient = await voteStakeRegistryClient_1.VsrClient.connect(provider, voteStakeRegistryClient_1.DEFAULT_VSR_ID);
    return vsrClient;
}
async function updateTokenParams() {
    const [client, wallet] = await Promise.all([buildClient(), setupWallet()]);
    const vsrClient = await setupVsr(client.connection, wallet);
    const group = await client.getGroup(constants_1.MANGO_V4_MAIN_GROUP);
    const instructions = [];
    const midPriceImpacts = (0, listingTools_1.getMidPriceImpacts)(group.pis);
    Array.from(group.banksMapByTokenIndex.values())
        .map((banks) => banks[0])
        .filter((bank) => bank.mint.toBase58() == 'So11111111111111111111111111111111111111112' ||
        bank.name.toLocaleLowerCase().indexOf('usdc') > -1 ||
        bank.name.toLocaleLowerCase().indexOf('stsol') > -1)
        .forEach(async (bank) => {
        // Limit borrows to 1/3rd of deposit, rounded to 1000, only update if more than 10% different
        const depositsInUsd = bank.nativeDeposits().mul(bank.price);
        let newNetBorrowLimitPerWindowQuote = depositsInUsd.toNumber() / 3;
        newNetBorrowLimitPerWindowQuote =
            Math.round(newNetBorrowLimitPerWindowQuote / 1_000_000_000) *
                1_000_000_000;
        newNetBorrowLimitPerWindowQuote =
            Math.abs((newNetBorrowLimitPerWindowQuote -
                bank.netBorrowLimitPerWindowQuote.toNumber()) /
                bank.netBorrowLimitPerWindowQuote.toNumber()) > 0.1
                ? newNetBorrowLimitPerWindowQuote
                : null;
        // Kick in weight scaling as late as possible until liquidation fee remains reasonable
        // Only update if more than 10% different
        let newWeightScaleQuote = null;
        if (bank.tokenIndex != 0 && // USDC
            bank.mint.toBase58() != 'So11111111111111111111111111111111111111112' // SOL
        ) {
            const PRESETS = bank?.oracleProvider === oracle_1.OracleProvider.Pyth
                ? listingTools_1.LISTING_PRESETS_PYTH
                : listingTools_1.LISTING_PRESETS;
            const tokenToPriceImpact = midPriceImpacts
                .filter((x) => x.avg_price_impact_percent < 1)
                .reduce((acc, val) => {
                if (!acc[val.symbol] ||
                    val.target_amount > acc[val.symbol].target_amount) {
                    acc[val.symbol] = val;
                }
                return acc;
            }, {});
            const priceImpact = tokenToPriceImpact[getApiTokenName(bank.name)];
            const suggestedTier = (0, listingTools_1.getProposedTier)(PRESETS, priceImpact?.target_amount, bank.oracleProvider === oracle_1.OracleProvider.Pyth);
            newWeightScaleQuote =
                PRESETS[suggestedTier].borrowWeightScaleStartQuote;
            newWeightScaleQuote =
                bank.depositWeightScaleStartQuote !== newWeightScaleQuote ||
                    bank.borrowWeightScaleStartQuote !== newWeightScaleQuote
                    ? newWeightScaleQuote
                    : null;
        }
        if (newNetBorrowLimitPerWindowQuote == null &&
            newWeightScaleQuote == null) {
            return;
        }
        const params = (0, builder_1.Builder)(clientIxParamBuilder_1.NullTokenEditParams)
            .netBorrowLimitPerWindowQuote(newNetBorrowLimitPerWindowQuote)
            .borrowWeightScaleStartQuote(newWeightScaleQuote)
            .depositWeightScaleStartQuote(newWeightScaleQuote)
            .build();
        const ix = await client.program.methods
            .tokenEdit(params.oracle, params.oracleConfig, params.groupInsuranceFund, params.interestRateParams, params.loanFeeRate, params.loanOriginationFeeRate, params.maintAssetWeight, params.initAssetWeight, params.maintLiabWeight, params.initLiabWeight, params.liquidationFee, params.stablePriceDelayIntervalSeconds, params.stablePriceDelayGrowthLimit, params.stablePriceGrowthLimit, params.minVaultToDepositsRatio, params.netBorrowLimitPerWindowQuote !== null
            ? new anchor_2.BN(params.netBorrowLimitPerWindowQuote)
            : null, params.netBorrowLimitWindowSizeTs !== null
            ? new anchor_2.BN(params.netBorrowLimitWindowSizeTs)
            : null, params.borrowWeightScaleStartQuote, params.depositWeightScaleStartQuote, params.resetStablePrice ?? false, params.resetNetBorrowLimit ?? false, params.reduceOnly, params.name, params.forceClose, params.tokenConditionalSwapTakerFeeRate, params.tokenConditionalSwapMakerFeeRate, params.flashLoanDepositFeeRate)
            .accounts({
            group: group.publicKey,
            oracle: bank.oracle,
            admin: group.admin,
            mintInfo: group.mintInfosMapByTokenIndex.get(bank.tokenIndex)
                ?.publicKey,
        })
            .remainingAccounts([
            {
                pubkey: bank.publicKey,
                isWritable: true,
                isSigner: false,
            },
        ])
            .instruction();
        const tx = new web3_js_1.Transaction({ feePayer: wallet.publicKey }).add(ix);
        const simulated = await client.connection.simulateTransaction(tx);
        if (simulated.value.err) {
            console.log('error', simulated.value.logs);
            throw simulated.value.logs;
        }
        console.log(`Bank ${bank.name}`);
        console.log(`- netBorrowLimitPerWindowQuote UI old ${(0, utils_1.toUiDecimalsForQuote)(bank.netBorrowLimitPerWindowQuote.toNumber()).toLocaleString()} new ${(0, utils_1.toUiDecimalsForQuote)(newNetBorrowLimitPerWindowQuote).toLocaleString()}`);
        console.log(`- WeightScaleQuote UI old ${(0, utils_1.toUiDecimalsForQuote)(bank.depositWeightScaleStartQuote).toLocaleString()} new ${(0, utils_1.toUiDecimalsForQuote)(newWeightScaleQuote).toLocaleString()}`);
        instructions.push(ix);
    });
    const tokenOwnerRecordPk = await (0, spl_governance_1.getTokenOwnerRecordAddress)(constants_2.MANGO_GOVERNANCE_PROGRAM, constants_2.MANGO_REALM_PK, constants_2.MANGO_MINT, new web3_js_1.PublicKey(VSR_DELEGATE_FROM_PK));
    const [tokenOwnerRecord, proposals] = await Promise.all([
        (0, spl_governance_1.getTokenOwnerRecord)(client.connection, tokenOwnerRecordPk),
        (0, spl_governance_1.getAllProposals)(client.connection, constants_2.MANGO_GOVERNANCE_PROGRAM, constants_2.MANGO_REALM_PK),
    ]);
    const walletSigner = wallet;
    if (!DRY_RUN) {
        const proposalAddress = await (0, createProposal_1.createProposal)(client.connection, walletSigner, constants_2.MANGO_DAO_WALLET_GOVERNANCE, tokenOwnerRecord, PROPOSAL_TITLE ? PROPOSAL_TITLE : 'Update risk parameters for tokens', '', Object.values(proposals).length, instructions, vsrClient);
        console.log(proposalAddress.toBase58());
    }
}
async function main() {
    try {
        await updateTokenParams();
    }
    catch (error) {
        console.log(error);
    }
}
try {
    main();
}
catch (error) {
    console.log(error);
}
